/**
 * @constructor Apemandocker
 * @param {object} config
 */
'use strict'

const execcli = require('execcli')
const co = require('co')
const os = require('os')
const childProcess = require('child_process')
const Promise = require('apemanpromise')
const argx = require('argx')
const stringcase = require('stringcase')

/** @lends Apemandocker */
function Apemandocker (config) {
  const s = this
  Object.assign(s, config || {})
}

Apemandocker.prototype = {
  _docker (args) {
    return execcli('docker', args, {
      notfound: 'install docker from https://docs.docker.com'
    })
  },
  _timeout (value, delay) {
    return new Promise((resolve) =>
      setTimeout(() => resolve(value), delay)
    )
  },
  /**
   * Build docker source.
   * @param {string} source - Name of source.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  buildImage (source, options = {}) {
    const s = this
    return s._docker([ 'build', options, source ])
  },
  /**
   * Remove docker image.
   * @param {string} image - Name of image.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  removeImage (image, options = {}) {
    const s = this
    return s._docker([ 'rmi', options, image ])
  },
  /**
   * List images.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  listImages (options = {}) {
    const s = this

    return co(function * () {
      let images = yield new Promise((resolve, reject) =>
        childProcess.exec(`docker images`, (err, stdOut, stdErr) =>
          err ? reject(err) : resolve({
            message: String(stdOut).trim(),
            errMessage: String(stdErr).trim()
          })
        )
      )
      if (images.errMessage) {
        console.error(images.errMessage)
      }
      let lines = images.message.split(os.EOL)
      let headers = lines.shift().split(/\s\s+/g).map((key) =>
        stringcase.camelcase(String(key).trim().replace(/\s/g, '_'))
      )
      return lines
        .filter((line) => !!line)
        .map((line) => {
          let data = {}
          line.split(/\s\s+/g).forEach((val, i) => {
            let key = headers[ i ]
            if (key) {
              data[ key ] = val
            }
          })
          return data
        })
    })
  },
  /**
   * Check if image exists.
   * @param {string} image - Name of image.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  imageExists (image, options = {}) {
    const s = this
    return co(function * () {
      let list = yield s.listImages({})
      let activeList = list.filter((filtering) => {
        let repository = String(filtering.repository).trim()
        let tag = String(filtering.tag).trim()
        return (repository === image) || ([ repository, tag ].join(':') === image)
      })
      return activeList.length > 0
    })
  },
  /**
   * Create a container.
   * @param {string} image - Name of image.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  createContainer (image, options = {}) {
    const s = this
    return s._docker([ 'create', options, image ])
  },
  /**
   * Execute a command inside a container.
   * @param {string} container - Name of container.
   * @param {string} script - Script to execute.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  execInContainer (container, script, options = {}) {
    const s = this
    return s._docker([ 'exec', options, container ].concat(script))
  },
  /**
   * Remove a container.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  removeContainer (container, options = {}) {
    const s = this
    let args = argx(arguments)
    options = args.pop('object') || {}
    return s._docker([ 'rm', options, container ])
  },
  /**
   * Start a container.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  startContainer (container, options = {}) {
    const s = this
    return co(function * () {
      let result = yield s._docker([ 'start', options, container ])
      return s._timeout(result, 10)
    })
  },
  /**
   * Run a container.
   * @param {string} image - Name of image.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  runContainer (image, options = {}) {
    const s = this
    return s._docker([ 'run', options, image ])
  },
  /**
   * Stop a container.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  stopContainer (container, options = {}) {
    const s = this
    return co(function * () {
      let result = yield s._docker([ 'stop', options, container ])
      return s._timeout(result, 10)
    })
  },
  /**
   * List containers
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  listContainers (options = {}) {
    const s = this
    let all = !!options.all

    const keys = [
      'ID',
      'Image',
      'Command',
      'CreatedAt',
      'RunningFor',
      'Ports',
      'Status',
      'Size',
      'Names',
      'Labels'
    ]
    let joiner = '|'

    return co(function * () {
      let format = keys.map((key) => {
        return `{ {.${key} }}`
      }).join(joiner)
      let containers = yield new Promise((resolve, reject) =>
        childProcess.exec(`docker ps${all ? ' -a' : ''} --format "${format}"`, (err, stdOut, stdErr) =>
          err ? reject(err) : resolve({
            message: String(stdOut).trim(),
            errMessage: String(stdErr).trim()
          })
        )
      )
      if (containers.errMessage) {
        console.error(containers.errMessage)
      }
      return containers.message.trim().split(os.EOL)
        .filter((line) => !!String(line).trim())
        .map((line) => {
          let data = {}
          String(line).split(joiner).forEach((value, i) => {
            let key = stringcase.camelcase(String(keys[ i ]).trim().replace(/\s/g, '_'))
            data[ key ] = String(value).trim()
          })
          return data
        })
    })
  },
  /**
   * Check if container exists.
   * @param {string} container - Name of container.
   * @param options
   * @returns {Promise}
   */
  containerExists (container, options = {}) {
    const s = this
    return co(function * () {
      let list = yield s.listContainers({
        all: true
      })
      return list.filter((filtering) =>
          String(filtering.names).replace(/,/g, '') === container
        ).length > 0
    })
  },
  /**
   * Detect a container running.
   * @param {string} container - Name of container.
   * @returns {Promise}
   */
  isContainerRunning (container) {
    let command = `docker inspect -f {{.State.Running}} ${container}`

    return new Promise((resolve, reject) =>
      childProcess.exec(command, (err, stdOut, stdErr) => {
        let isRunning = !err && (String(stdOut).trim().toLowerCase() === "true")
        resolve(isRunning)
      })
    )
  },
  /**
   * Inspect a container.
   * @param container
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  inspectContainer (container, options = {}) {
    const s = this
    let command = `docker inspect ${container}`
    return new Promise((resolve, reject) =>
      childProcess.exec(command, (err, stdOut, stdErr) => {
        let result = !err && JSON.parse(String(stdOut).trim())
        resolve(err, result)
      })
    )
  },
  /**
   * Log container.
   * @param container
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  logContainer (container, options = {}) {
    const s = this
    return s._docker([ 'logs', options, container ])
  },
  /**
   * Check if network exists.
   * @param network
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  networkExists (network, options = {}) {
    const s = this

    return co(function * () {
      let networks = yield s.listNetworks(options)
      for (let i = 0, len = networks.length; i < len; i++) {
        let hit = networks[ i ].name === network
        if (hit) {
          return true
        }
      }
      return false
    })
  },
  /**
   * Open terminal.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @param {string} [options.shell='bash'] - Shell for terminal
   */
  terminal (container, options = {}) {
    const s = this
    return co(function * () {
      return childProcess.spawn('docker',
        [ 'exec', '-it', container, options.shell || 'bash' ], {
          stdio: 'inherit'
        }
      )
    })
  },
  /**
   * Create a network.
   * @param {string} network - Name of network.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  createNetwork (network, options = {}){
    const s = this
    return s._docker([ 'network', 'create', options, network ])
  },
  /**
   * Connect network.
   * @param {string} network - Name of network.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  connectNetwork (network, container, options = {}){
    const s = this
    return s._docker([ 'network', 'connect', options, network, container ])
  },
  /**
   * Disconnect network.
   * @param {string} network - Name of network.
   * @param {string} container - Name of container.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  disconnectNetwork (network, container, options = {}) {
    const s = this
    return s._docker([ 'network', 'disconnect', options, network, container ])
  },
  /**
   * List net works.
   * @param {object} [options] - Optional settings.
   * @returns {Promise}
   */
  listNetworks (options = {}) {
    const s = this
    let command = 'docker network ls'
    return co(function * () {
      let networks = yield new Promise((resolve, reject) =>
        childProcess.exec(command, (err, stdOut, stdErr) =>
          err ? reject(err) : resolve({
            errMessage: String(stdErr).trim(),
            message: String(stdOut).trim()
          })
        )
      )
      if (networks.errMessage) {
        console.error(networks.errMessage)
      }
      let lines = networks.message.split(os.EOL)

      let headers = lines.shift().split(/\s\s+/g).map((key) =>
        stringcase.camelcase(String(key).trim().replace(/\s/g, '_'))
      )

      return lines.map((line) => {
        let data = {}
        line.split(/\s\s+/g).forEach((val, i) => {
          let key = headers[ i ]
          if (key) {
            data[ key ] = val
          }
        })
        return data
      })
    })
  },
  removeNetwork (name, options = {}) {
    const s = this
    return s._docker([ 'network', 'rm', options, name ])
  }
}

module.exports = Apemandocker
